# ============================================================================ #
#   KML Document templates
# ============================================================================ #

document_kml_template = """<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2" xmlns:gx="http://www.google.com/kml/ext/2.2" xmlns:kml="http://www.opengis.net/kml/2.2" xmlns:atom="http://www.w3.org/2005/Atom">
<Document>
    <name>$name</name>
    <StyleMap id="vesselStyleMap">
        <Pair>
            <key>normal</key>
            <styleUrl>#normVesselStyle</styleUrl>
        </Pair>
        <Pair>
            <key>highlight</key>
            <styleUrl>#hlightVesselStyle</styleUrl>
        </Pair>
    </StyleMap>
    <Style id="hlightVesselStyle">
        <IconStyle>
            <scale>1.2</scale>  
            <color>ff006666</color>
            <Icon>
                <href>$icon</href>
            </Icon>
            <hotSpot x="16" y="3" xunits="pixels" yunits="pixels"/>
        </IconStyle>
        <LabelStyle><color>ff999999</color><scale>0.8</scale></LabelStyle>
    </Style>
    <Style id="normVesselStyle">
        <IconStyle>
          <color>ff009999</color>
          <Icon><href>$icon</href></Icon>
          <hotSpot x="16" y="3" xunits="pixels" yunits="pixels"/>
        </IconStyle>
        <LabelStyle><scale>0</scale></LabelStyle>
    </Style>
    <StyleMap id="trackStyle1">
        <Pair>
            <key>normal</key>
            <styleUrl>#normTrackStyle1</styleUrl>
        </Pair>
        <Pair>
            <key>highlight</key>
            <styleUrl>#hlightTrackStyle1</styleUrl>
        </Pair>
    </StyleMap>
    <Style id="hlightTrackStyle1">
        <LineStyle>
            <color>ff999999</color>
            <width>2</width>
        </LineStyle>
    </Style>
    <Style id="normTrackStyle1">
        <LineStyle>
            <color>ff999999</color>
            <width>1.2</width>
        </LineStyle>
    </Style>    
    <StyleMap id="trackStyle2">
        <Pair>
            <key>normal</key>
            <styleUrl>#normTrackStyle2</styleUrl>
        </Pair>
        <Pair>
            <key>highlight</key>
            <styleUrl>#hlightTrackStyle2</styleUrl>
        </Pair>
    </StyleMap>
    <Style id="hlightTrackStyle2">
        <LineStyle>
            <color>ff00ffff</color>
            <width>6</width>
        </LineStyle>
    </Style>
    <Style id="normTrackStyle2">
        <LineStyle>
            <color>ff00ffff</color>
            <width>4</width>
        </LineStyle>
    </Style>    
    <StyleMap id="trackStyle3">
        <Pair>
            <key>normal</key>
            <styleUrl>#normTrackStyle3</styleUrl>
        </Pair>
        <Pair>
            <key>highlight</key>
            <styleUrl>#hlightTrackStyle3</styleUrl>
        </Pair>
    </StyleMap>
    <Style id="hlightTrackStyle3">
        <LineStyle>
            <color>ff0000ff</color>
            <width>6</width>
        </LineStyle>
    </Style>
    <Style id="normTrackStyle3">
        <LineStyle>
            <color>ff0000ff</color>
            <width>4</width>
        </LineStyle>
    </Style>        
    <StyleMap id="trackStyle4">
        <Pair>
            <key>normal</key>
            <styleUrl>#normTrackStyle4</styleUrl>
        </Pair>
        <Pair>
            <key>highlight</key>
            <styleUrl>#hlightTrackStyle4</styleUrl>
        </Pair>
    </StyleMap>
    <Style id="hlightTrackStyle4">
        <LineStyle>
            <color>00000000</color>
            <width>6</width>
        </LineStyle>
    </Style>
    <Style id="normTrackStyle4">
        <LineStyle>
            <color>00000000</color>
            <width>4</width>
        </LineStyle>
    </Style>
    $track_kml
</Document>
</kml>
"""


# ============================================================================ #
#   Vessel template
# ============================================================================ #

vessel_kml_template = """
<Folder>
    <name>Vessel Tracks</name>
    $track_kml
</Folder>

<Folder>
    <name>AIS Points</name>
    $placemarks_kml
</Folder>
"""


# ============================================================================ #
#   Track template
# ============================================================================ #

track_template = """    <Placemark>
    <name>$name</name>
    <styleUrl>#$style</styleUrl>
    <LineString>
        <tessellate>1</tessellate>
        <coordinates>
            $coords
        </coordinates>
    </LineString>
    <TimeSpan><begin>$time_begin</begin><end>$time_end</end></TimeSpan>
     
</Placemark>"""


# ============================================================================ #
#   Placemark folder template
# ============================================================================ #

placemark_folder_template = """<Folder>
    <name>$name</name>
$placemarks_kml
</Folder>
"""


# ============================================================================ #
#   Placemark template
# ============================================================================ #

placemark_template = """<Placemark>
    <name>$timestamp</name>
    <description>
<table width="300">
    <tr><th align="right">Vessel Name</th><td>$name</td></tr>
    <tr><th align="right">Vessel Type</th><td>$shiptype</td></tr>
    <tr><th align="right">MMSI</th><td>$mmsi</td></tr>
    <tr><th align="right">Length</th><td>$to_bow</td></tr>
    <tr><th align="right">Datetime</th><td>$timestamp</td></tr>
    <tr><th align="right">True Heading</th><td>$heading</td></tr>
    <tr><th align="right">Speed Over Ground</th><td>$speed</td></tr>
    <tr><th align="right">Course Over Ground</th><td>$course</td></tr>
    <tr><th align="right">Latitude</th><td>$lat</td></tr>
    <tr><th align="right">Longitude</th><td>$lon</td></tr>
    <tr><th align="right">Vessel Info</th><td><a href="$marinetraffic_url">marinetraffic.com</a> <a href="$itu_url">ITU</a></td></tr>
</table>    
    </description>    
    <styleUrl>#vesselStyleMap</styleUrl>
    <TimeStamp><when>$when</when></TimeStamp>
    <Point>
        <coordinates>$lon,$lat,0</coordinates>
    </Point>
    <Style><IconStyle><heading>$course</heading><color>$icon_color</color></IconStyle></Style>
</Placemark>"""



import kmltrack.generator
import os.path
import dateutil.parser
import csv
import json
try:
    import msgpack
except:
    pass

def column_mapper(rows, **mapping):
    "Usage: column_mapper(rows, dst_col1=src_col1, dst_col2=src_col2, ...)"
    env = {
        "d": dateutil.parser.parse
        }
    for row in rows:
        renv = dict(row)
        renv.update(env)
        renv['row'] = row
        for key, expr in mapping.iteritems():
            try:
                row[key] = eval(expr, renv)
            except:
                pass
        yield row

def file_to_kml(infile_name, outfile_name, reader, **kw):
    column_map = {
        'timestamp': 'd(timestamp)',
        'lat': 'float(lat)',
        'lon': 'float(lon)',
        'color': 'float(color)',
        'course': 'float(course)'
        }
    column_map.update(kw.get('column_map', {}))

    def input():
        with sys.stdin if infile_name is None or '-' == infile_name else open(infile_name, 'rb') as in_file:
            for row in column_mapper(reader(in_file), **column_map):
                yield row

    with sys.stdout if outfile_name is None or '-' == outfile_name else open(outfile_name, 'w') as kml_file:
        kw['input'] = input
        kw["name"] = os.path.splitext(outfile_name)[0]
        kmltrack.generator.DocumentKMLTemplate(kml_file, kw)


def csv_to_kml(infile_name, outfile_name, **kw):
     file_to_kml(infile_name, outfile_name, cvs.DictReader, **kw)

def json_to_kml(infile_name, outfile_name, **kw):
    def reader(file):
        for row in file:
            yield json.loads(row)
    file_to_kml(infile_name, outfile_name, reader, **kw)

def msgpack_to_kml(infile_name, outfile_name, **kw):
    file_to_kml(infile_name, outfile_name, msgpack.Unpacker, **kw)



import re
import types

class Template(object):
    """A generic templating system using the same syntax as Pythons string
    Template, but generates output directly to a file, without
    constructing the whole thing in memory first.
    """

    template = ""
    ignore_missing = False

    def __init__(self, *arg, **kw):
        if arg or kw:
            self.render(*arg, **kw)

    def render(self, out_file, context=None):
        if context is None:
            context = {}
        for escape, parenvar, var, verbatim in re.findall(r"([$][$])|([$]{[a-z0-9_]*})|([$][a-z0-9_]*)|([^$]*)", self.template):
            if escape:
                out_file.write('$')
            elif parenvar:
                self.writevar(parenvar[2:-1], out_file, context)
            elif var:
                self.writevar(var[1:], out_file, context)
            else:
                out_file.write(verbatim)

    def writevar(self, name, out_file, context):
        cls = type(self)
        if name in context:
            value = context[name]
        elif hasattr(self, name):
            value = getattr(self, name)
        elif hasattr(cls, name):
            value = getattr(cls, name)
        elif self.ignore_missing:
            return
        else:
            raise Exception("Unknown substitution %s not in %s" % (name, context.keys()))

        if isinstance(value, type) and issubclass(value, Template):
            value(out_file, context)
        elif isinstance(value, (types.MethodType, types.FunctionType)):
            value(out_file, context)
        else:
            out_file.write(unicode(value).encode("utf-8"))



import sys
import math
import kmltrack.template
import kmltrack.iterators
import kmltrack.template_strings

class DocumentKMLTemplate(kmltrack.template.Template):
    template = kmltrack.template_strings.document_kml_template
    icon = 'http://alerts.skytruth.org/markers/triangle-up.png'
    name = 'out'

    class track_kml(kmltrack.template.Template):
        template = kmltrack.template_strings.vessel_kml_template

        @staticmethod
        def rows(context):
            # Note: this method is called twice times!

            context['reader'] = context['input']()

            for row in context['reader']:
                if 'timestamp' not in row or 'lat' not in row or 'lon' not in row:
                    if context.get('verify-rows', False):
                        raise Exception("Missing column (lat, lon, timestamp): " + str(row))
                    continue
                mmsi = row.get('mmsi', '')
                if not row.get('name', ''):
                    row['name'] = mmsi
                row['marinetraffic_url'] = 'http://www.marinetraffic.com/ais/shipdetails.aspx?MMSI=%s'% mmsi
                row['itu_url'] = 'http://www.itu.int/cgi-bin/htsh/mars/ship_search.sh?sh_mmsi=%s' % mmsi

                ctx = dict(context)
                ctx.update(row)
                yield ctx

        class PlacemarkFolderTemplate(kmltrack.template.Template):
            template = kmltrack.template_strings.placemark_folder_template

            class PlacemarkTemplate(kmltrack.template.Template):
                template = kmltrack.template_strings.placemark_template
                ignore_missing = True

                @staticmethod
                def when(out_file, context):
                    out_file.write(context['timestamp'].strftime('%Y-%m-%dT%H:%M:%SZ'))

                @staticmethod
                def icon_color(out_file, context):
                    c = context.get('color', 0)

                    if isinstance(c, (float, int, long)):
                        def fudge(v):
                            a = 0.1
                            return v/2.0 - math.sqrt(4.0*a*a + v*v - 2.0*v + 1.0) / 2.0 + 1.0/2.0
                        c = (
                            255, # Alpha
                            fudge(c * 1.0) * 255, # Blue
                            fudge(c * 1.5) * 255, # Green
                            fudge(c * 3.0) * 255 # Red
                            )

                    if isinstance(c, tuple):
                        c = '%02x%02x%02x%02x' % c
                    
                    out_file.write(c)

            def placemarks_kml(self, out_file, context):
                for row in context['rows']:
                    ctx = dict(context)
                    ctx.update(row)
                    self.PlacemarkTemplate(out_file, ctx)
                    if not hasattr(context['rows'], 'peek'):
                        return
                    next_date = context['rows'].peek['timestamp'].date()
                    if next_date != context['prev_date']:
                        return

        def placemarks_kml(self, out_file, context):
            context['rows'] = kmltrack.iterators.lookahead(self.rows(context))

            while hasattr(context['rows'], 'peek'):
                context['prev_date'] = context['rows'].peek['timestamp'].date()
                context['name'] = context['prev_date'].strftime('%Y-%m-%d')
                self.PlacemarkFolderTemplate(out_file, context)

        time_gap_styles = [
            {'range': (0, (3600 * 12)), 'style': 'trackStyle1'},
            {'range': ((3600 * 12), (3600 * 48)), 'style': 'trackStyle2'},
            {'range': ((3600 * 48), (3600 * 168)), 'style': 'trackStyle3'},
            {'range': ((3600 * 168), sys.maxint), 'style': 'trackStyle4'},
            ]

        def get_time_gap_style(self, dt, last_dt):
            td = (dt - last_dt).total_seconds()

            for s in self.time_gap_styles:
                if td >= s['range'][0] and td < s['range'][1]:
                    return s['style']

        def row_pairs(self, context):
            last = None
            for row in self.rows(context):
                if last is not None:
                    yield {
                        'src': last,
                        'dst': row,
                        'style': self.get_time_gap_style(row['timestamp'], last['timestamp'])
                        }
                last = row

        class TrackSegmentTemplate(kmltrack.template.Template):
            template = kmltrack.template_strings.track_template
            name = 'Vessel Track'

            @staticmethod
            def coords(out_file, context):
                first = True
                for row_pair in context['row_pairs']:
                    context['time_end'] = row_pair['dst']['timestamp'].strftime('%Y-%m-%dT%H:%M:%SZ')

                    if first:
                        out_file.write('%(lon)s,%(lat)s,0 ' % row_pair['src'])
                        first = False
                    out_file.write('%(lon)s,%(lat)s,0 ' % row_pair['dst'])

                    if (not hasattr(context['row_pairs'], 'peek')
                        or context['row_pairs'].peek['style'] != row_pair['style']):
                        break

        def track_kml(self, out_file, context):
            context['row_pairs'] = kmltrack.iterators.lookahead(self.row_pairs(context))

            first = True
            while hasattr(context['row_pairs'], 'peek'):
                next_row_pair = context['row_pairs'].peek
                if first:
                    context['time_begin'] = next_row_pair['src']['timestamp'].strftime('%Y-%m-%dT%H:%M:%SZ')
                    first = False
                context['style'] = next_row_pair['style']

                self.TrackSegmentTemplate(out_file, context)



class lookahead(object):
    """An iterator wrapper that allows peeking at the next value without consuming it."""
    def __init__(self, it):
        self.it = iter(it)
        try:
            self.peek = self.it.next()
        except StopIteration:
            pass
    def __iter__(self):
        return self
    def next(self):
        try:
            return self.peek
        except:
            raise StopIteration
        finally:
            try:
                self.peek = self.it.next()
            except StopIteration:
                if hasattr(self, 'peek'):
                    del self.peek



#! /usr/bin/python

from setuptools.command import easy_install
from setuptools import setup, find_packages
import shutil
import os.path
import sys
import hashlib

PKG_DIR = os.path.abspath(os.path.dirname(__file__))
PKG_NAME = os.path.basename(PKG_DIR)

# Make it possible to overide script wrapping
old_is_python_script = easy_install.is_python_script
def is_python_script(script_text, filename):
    if 'SETUPTOOLS_DO_NOT_WRAP' in script_text:
        return False
    return old_is_python_script(script_text, filename)
easy_install.is_python_script = is_python_script

setup(
    name = "kmltrack",
    description = "kmltrack converts a track (a series of lat/lon/timestamp and optional measurement values) into a kml file.",
    keywords = "kml",
    install_requires = ["python-dateutil"],
    version = "0.0.2",
    author = "Egil Moeller",
    author_email = "egil@skytruth.org",
    license = "GPL",
    url = "https://github.com/SkyTruth/kmltrack",
    packages=[
        'kmltrack',
    ],
    scripts = ["bin/kmltrack"]
)
